static const FFFilter * const filter_list[] = {
    &ff_vf_bwdif,
    &ff_vf_colorspace_cuda,
    &ff_vf_crop,
    &ff_vf_deblock,
    &ff_vf_format,
    &ff_vf_hflip,
    &ff_vf_hwdownload,
    &ff_vf_hwmap,
    &ff_vf_hwupload,
    &ff_vf_monochrome,
    &ff_vf_pad,
    &ff_vf_scale,
    &ff_vf_scale_cuda,
    &ff_vf_tonemap,
    &ff_vf_transpose,
    &ff_vf_vflip,
    &ff_vf_vpp_qsv,
    &ff_vf_yadif,
    &ff_vf_zscale,
    &ff_asrc_abuffer,
    &ff_vsrc_buffer,
    &ff_asink_abuffer,
    &ff_vsink_buffer,
    NULL };
