static const FFOutputFormat * const muxer_list[] = {
    &ff_ass_muxer,
    &ff_ipod_muxer,
    &ff_matroska_muxer,
    &ff_mov_muxer,
    &ff_mp4_muxer,
    &ff_psp_muxer,
    &ff_srt_muxer,
    &ff_sup_muxer,
    &ff_webm_muxer,
    NULL };
